from collections import deque
import logging

import gevent

from .base import PoolSink
from ..asynchronous import AsyncResult
from ..constants import (Int, ChannelState, SinkProperties, SinkRole)
from ..sink import (
  ClientMessageSink,
  SinkProvider,
  FailingMessageSink
)
from ..dispatch import ServiceClosedError
from ..varz import (
  Gauge,
  Source,
  VarzBase
)

class QueuingMessageSink(ClientMessageSink):
  def __init__(self, queue):
    super(QueuingMessageSink, self).__init__()
    self._queue = queue

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    self._queue.append((sink_stack, msg, stream, headers))

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("This should never be called")

  @property
  def state(self):
    return ChannelState.Open


class MaxWaitersError(Exception): pass

class WatermarkPoolSink(PoolSink):
  """A watermark pool keeps a cached number of sinks active (the low watermark).
  Once the low watermark is hit, the pool will create new sinks until it hits the
  high watermark.  At that point, it will begin queuing requests.

  The pool guarantees only a single request will be active on any underlying sink
  at any given time, that is, each sink processes requests serially.
  """

  ROOT_LOG = logging.getLogger('scales.pool.WatermarkPool')

  class Varz(VarzBase):
    """
    size - The current size of the pool.
    queue_size - The length of the waiter queue.
    min_size - The configured low-watermark.
    max_size - The configured high-watermark.
    """
    _VARZ_BASE_NAME = 'scales.pool.WatermarkPool'
    _VARZ = {
      'size': Gauge,
      'queue_size': Gauge
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    endpoint = global_properties[SinkProperties.Endpoint]
    name = global_properties[SinkProperties.Label]

    self._cache = deque()
    self._waiters = deque()
    self._min_size = sink_properties.min_watermark
    self._max_size = sink_properties.max_watermark
    self._max_queue_size = sink_properties.max_queue_len
    self._current_size = 0
    self._state = ChannelState.Idle
    socket_name = '%s:%s' % (endpoint.host, endpoint.port)
    self.endpoint = socket_name
    self._varz = self.Varz(Source(service=name,
                                  endpoint=socket_name))
    self._log = self.ROOT_LOG.getChild('[%s.%s]' % (name, socket_name))
    super(WatermarkPoolSink, self).__init__(next_provider, global_properties)

  def __PropagateShutdown(self, value):
    self.on_faulted.Set(value)

  def _DiscardSink(self, sink):
    """Close the sink and unsubscribe from fault notifications

    Args:
      sink - The sink to discard.
    """
    sink.on_faulted.Unsubscribe(self.__PropagateShutdown)
    sink.Close()

  def _Dequeue(self):
    """Attempt to get a sink from the cache.

    Returns:
      A sink if one can be taken from the cache, else None.
    """
    while any(self._cache):
      item = self._cache.popleft()
      if item.state <= ChannelState.Open:
        return item
      else:
        self._DiscardSink(item)
    return None

  def _Get(self):
    cached = self._Dequeue()
    if cached:
      return cached
    elif self._current_size < self._max_size:
      self._current_size += 1
      self._varz.size(self._current_size)
      sink = self._sink_provider.CreateSink(self._properties)
      # TODO: we could get a better failure case here by detecting that Open()
      # failed and retrying, however for now the simplest option is to just fail.
      sink.Open().wait()
      sink.on_faulted.Subscribe(self.__PropagateShutdown)
      return sink
    else:
      if len(self._waiters) + 1 > self._max_queue_size:
        return FailingMessageSink(MaxWaitersError())
      else:
        self._varz.queue_size(len(self._waiters) + 1)
        return QueuingMessageSink(self._waiters)

  def _Release(self, sink):
    # Releasing a queuing sink is a noop
    if (isinstance(sink, QueuingMessageSink) or
        isinstance(sink, FailingMessageSink)):
      self._varz.queue_size(len(self._waiters))
      return

    do_close = False
    # This sink is already shutting down
    if self.state == ChannelState.Closed:
      self._current_size -= 1
    # One of the underlying sinks failed, shut down
    elif sink.state == ChannelState.Closed:
      self._current_size -= 1
      self.Close()
    # There are some waiters queued, reuse this sink to process another request.
    elif any(self._waiters):
      gevent.spawn(self._ProcessQueue, sink)
    # We're below the min-size specified, cache this sink
    elif self._current_size <= self._min_size:
      self._cache.append(sink)
    # We're above the min-size, close the sink.
    else:
      self._current_size -= 1
      do_close = True

    self._varz.size(self._current_size)
    if do_close:
      self._DiscardSink(sink)

  def _ProcessQueue(self, sink):
    """Called as a continuation of an underlying sink completing.  Get the
    next waiter and use 'sink' to process it.

    Args:
      sink - An open sink.
    """
    sink_stack, msg, stream, headers = self._waiters.popleft()
    self._varz.queue_size(len(self._waiters))
    # The stack has a QueuingChannelSink on the top now, pop it off
    # and push the real stack back on.
    orig_sink, ctx = sink_stack.Pop()
    sink_stack.Push(orig_sink, sink)
    sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def Open(self):
    ar = AsyncResult()
    ar.SafeLink(self._OpenImpl)
    return ar

  def _OpenImpl(self):
    sink = self._Get()
    self._Release(sink)
    self._state = ChannelState.Open

  def _FlushCache(self):
    [self._DiscardSink(sink) for sink in self._cache]

  def Close(self):
    self._state = ChannelState.Closed
    self._FlushCache()
    fail_sink = FailingMessageSink(ServiceClosedError)
    [fail_sink.AsyncProcessRequest(sink_stack, msg, stream, headers)
     for sink_stack, msg, stream, headers in self._waiters]

  @property
  def state(self):
    return self._state


WatermarkPoolSink.Builder = SinkProvider(
  WatermarkPoolSink,
  SinkRole.Pool,
  min_watermark = 1,
  max_watermark = Int.MaxValue,
  max_queue_len = Int.MaxValue
)



from setuptools import setup

INSTALL_REQUIRES = (
  'gevent>=1.3.0',
  'thrift>=0.10.0',
  'kazoo>=2.5.0',
  'six>=1.13.0',
  'requests>=2.0.0'
)

TESTS_REQUIRES = (
  'responses>=0.10.9',
)

setup(
  name='scales-rpc',
  version='2.0.1',
  author='Steve Niemitz',
  author_email='sniemitz@twitter.com',
  url='https://www.github.com/steveniemitz/scales',
  description='A python RPC client stack',
  summary='A generic python RPC client framework.',
  license='MIT License',
  packages=['scales',
            'scales.http',
            'scales.kafka',
            'scales.loadbalancer',
            'scales.mux',
            'scales.pool',
            'scales.redis',
            'scales.thrift',
            'scales.thrifthttp',
            'scales.thriftmux'],
  install_requires=INSTALL_REQUIRES,
  extras_require={
    'tests': INSTALL_REQUIRES + TESTS_REQUIRES
  }
)



import gevent

class Observable(object):
  """An object to which consumers can subscribe to notifications for its value
  changing."""
  __slots__ = ('_callbacks', '_one_shot_callbacks', '_value')

  def __init__(self):
    self._callbacks = None
    self._one_shot_callbacks = None
    self._value = None

  def __Notify(self, value):
    if self._callbacks:
      cb_copy = self._callbacks.copy()
      [cb(value) for cb in cb_copy]
    if self._one_shot_callbacks:
      cb_copy, self._one_shot_callbacks = self._one_shot_callbacks, None
      [cb(value) for cb in cb_copy]

  def Get(self):
    """Get the current value of the Observable."""
    return self._value

  def Set(self, value=None):
    """Set the current value of the Observable and notify any subscribers."""
    self._value = value
    gevent.spawn(self.__Notify, value)

  def Subscribe(self, callback, one_shot=False):
    """Subscribe to notifications of the value changing.

    Args:
      callback - The method to be called when the value changes.
      one_shot - If true, the callback will only be invoked once.
    """
    if callback is None or not callable(callback):
      raise Exception("callback can not be None")

    if one_shot:
      if not self._one_shot_callbacks:
        self._one_shot_callbacks = set()
      self._one_shot_callbacks.add(callback)
    else:
      if not self._callbacks:
        self._callbacks = set()
      self._callbacks.add(callback)

  def Unsubscribe(self, callback):
    """Unsubscribe from notifications.

    Args:
      callback - The method previously subscribed.
    """
    if self._callbacks:
      self._callbacks.discard(callback)
    if self._one_shot_callbacks:
      self._one_shot_callbacks.discard(callback)



import heapq
import logging
import math
import time

import gevent

from gevent.event import Event

LOG = logging.getLogger('scales.TimerQueue')


class LowResolutionTime(object):
  """Provides a low-resolution time source with significantly lower overhead
  than calling time.time()"""
  __slots__ = '_interval', 'now'

  def __init__(self, resolution=1):
    """
    Args:
      resolution - The resolution of the timer in seconds, defaults to 1 second.
    """
    self._interval = resolution
    self.now = time.time()
    self._Update()

  def _Update(self):
    last_time = self.now
    self.now = time.time()
    # Figure out how far we missed the interval by
    skew = max(0, (self.now - last_time) - self._interval)
    # Adjust the next tick by that skew
    next_tick = self.now + self._interval - skew
    GLOBAL_TIMER_QUEUE.Schedule(next_tick, self._Update)

  def Get(self):
    return self.now


class TimerQueue(object):
  """A timer that provides efficient scheduling of large numbers of events in
  the near future."""
  def __init__(self, time_source=time.time, resolution=0.01):
    """
    Args:
      time_source - A callable to get the current time (in seconds).
      resolution - The minimum resolution of the timer.  If set, all events are
                   quantized to this resolution.
    """
    self._queue = []
    self._event = Event()
    self._seq = 0
    self._resolution = resolution
    self._time_source = time_source
    self._worker = gevent.spawn(self._TimerWorker)

  def __del__(self):
    self._worker.kill(block=False)
    self._worker = None

  def _TimerWorker(self):
    while True:
      # If the queue is empty, wait for an item to be added.
      if not self._queue:
        self._event.wait()

      if self._event.is_set():
        self._event.clear()
        # A sleep here is needed to work around a bug with gevent.
        # If the event is cleared and then immediately waited on, the wait will
        # completed instantly and report it timed out.
        gevent.sleep(0)

      # Peek the head of the queue
      at, peeked_seq, cancelled = self._PeekNext()
      if cancelled:
        # The item has already been canceled, remove it and continue.
        heapq.heappop(self._queue)
        continue

      # Wait for a new item to be added to the queue,
      # or for the timeout to expire.
      to_wait = at - self._time_source()
      if to_wait > 0:
        # There's some time to wait, wait for it or a more recent item to be
        # added to the head of the queue.
        wait_timed_out = not self._event.wait(to_wait)
      else:
        # It should already be run, do it right now.
        wait_timed_out = True

      if wait_timed_out:
        # Nothing newer came in before it timed out.
        at, seq, cancelled, action = heapq.heappop(self._queue)
        # This is an assert that should never occur, if it does, we somehow
        # ran events out of order.
        if seq != peeked_seq:
          LOG.critical("seq != peeked_seq [%d, %d]" % (seq, peeked_seq))
        if not cancelled:
          # Run it
          gevent.spawn(action)
          # Clear the reference out in this loop
          del action
      else:
        # A newer item came in, nothing to do here, re-loop
        pass

  def _PeekNext(self):
    return self._queue[0][:3]

  def Schedule(self, deadline, action):
    """Schedule an operation

    Args:
      deadline - The absolute time this event should occur on.
      action - The action to run.
    Returns:
      A callable that can be invoked to cancel the scheduled operation.
    """
    if action is None:
      raise Exception("action must be non-null")

    if self._resolution:
      deadline = int(
        math.ceil(float(deadline) / self._resolution)) * self._resolution

    self._seq += 1
    timeout_args = [deadline, self._seq, False, action]
    
    def cancel():
      timeout_args[2] = True
      # Null out to avoid holding onto references.
      timeout_args[3] = None

    heapq.heappush(self._queue, timeout_args)
    # Wake up the waiter thread if this is now the newest
    if self._queue[0][0] == deadline:
      self._event.set()
    return cancel


GLOBAL_TIMER_QUEUE = TimerQueue()
LOW_RESOLUTION_TIME_SOURCE = LowResolutionTime()
LOW_RESOLUTION_TIMER_QUEUE = TimerQueue(
  time_source=LOW_RESOLUTION_TIME_SOURCE.Get,
  resolution=1)



"""Heap Load Balancer.

Based on the heap balancer from finagle, see (https://github.com/twitter/finagle/blob/master/finagle-core/src/main/scala/com/twitter/finagle/loadbalancer/HeapBalancer.scala)

The heap load balancer maintains all nodes in a min-heap.  The heap is adjusted
as the load on a node increases or decreases.  Nodes start at Zero (min int) and
increase load as they are used.  Load decreases when they are released back into
the pool.

Downed nodes are tracked by setting a node's load to > 0.  A linked list of downed
nodes is kept to resurrect downed nodes if they become active again.
"""

import functools
import random

try:
  from gevent.lock import RLock # pylint: disable=E0611
except ImportError:
  from gevent.coros import RLock

from .base import (
  LoadBalancerSink,
  NoMembersError
)
from ..asynchronous import AsyncResult
from ..constants import (
  ChannelState,
  Int,
  MessageProperties,
  SinkProperties,
  SinkRole
)
from ..sink import (
  FailingMessageSink,
  SinkProvider,
)
from ..varz import (
  Counter,
  Gauge,
  Source,
  VarzBase
)

class Heap(object):
  """A utility class to perform heap functions"""
  @staticmethod
  def Swap(heap, i, j):
    """Swap two elements in the heap.

    Args:
      heap - The heap array.
      i, j - The indexes int the array to swap.
    """
    if i == 0 or j == 0:
      raise Exception("heap swapping of element 0 should never happen.")

    heap[i], heap[j] = heap[j], heap[i]
    heap[i].index = i
    heap[j].index = j

  @staticmethod
  def FixUp(heap, i):
    """Traverse up the heap, ensuring the invariant is maintained.

    Args:
      heap - The heap array.
      i - The index to start at.
    """
    while True:
      if i != 1 and heap[i] < heap[i//2]:
        Heap.Swap(heap, i, i//2)
        i //= 2 # FixUp(heap, i/2)
      else:
        break

  @staticmethod
  def FixDown(heap, i, j):
    """Traverse down the heap, ensuring the invariant is maintained.

    Args:
      heap - The heap array.
      i, j - The node index to traverse from -> to.
    """
    while True:
      if j < i * 2: break

      m = 2 * i if (j == i * 2 or heap[2*i] < heap[2*i+1]) else 2*i+1
      if heap[m] < heap[i]:
        Heap.Swap(heap, i, m)
        i = m  # FixDown(heap, m, j)
      else:
        break


def synchronized(fn):
  """Runs the wrapped method under a lock.
  The self parameter to the wrapped function is expected to have a __heap_lock
  attribute.
  """
  @functools.wraps(fn)
  def wrapper(self, *args, **kwargs):
    with self._heap_lock:
      return fn(self, *args, **kwargs)
  return wrapper


class HeapBalancerSink(LoadBalancerSink):
  """A sink that implements a heap load balancer."""
  Penalty = Int.MaxValue
  Idle = Int.MinValue + 1

  class HeapVarz(VarzBase):
    """
    size - The number of nodes in the pool
    no_members - The number of times the balancer served a failing requests
                 because there were no members in the pool.
    """
    _VARZ_BASE_NAME = 'scales.loadbalancer.Heap'
    _VARZ = {
      'size': Gauge,
      'no_members': Counter
    }

  class Node(object):
    __slots__ = ('load', 'index', 'downq', 'avg_load', 'channel', 'endpoint')

    def __init__(self, channel, load, index, endpoint):
      self.channel = channel
      self.avg_load = 0
      self.load = load
      self.index = index
      self.downq = None
      self.endpoint = endpoint

    def __lt__(self, other):
      """Compare to other, return true if (load, index) < other.(load, index)"""
      if self.load > other.load:
        return False
      elif self.load < other.load:
        return True
      else:
        return self.index < other.index

  def __init__(self, next_provider, sink_properties, global_properties):
    self._heap = [self.Node(
        FailingMessageSink(functools.partial(Exception, "this sink should never be used")),
        self.Idle, 0, None)]
    self._no_members = FailingMessageSink(NoMembersError)
    self._downq = None
    self._size = 0
    self._open = False
    self._heap_lock = RLock()
    service_name = global_properties[SinkProperties.Label]
    self.__varz = self.HeapVarz(Source(service=service_name))
    super(HeapBalancerSink, self).__init__(next_provider, sink_properties, global_properties)

  def _AsyncProcessRequestImpl(self, sink_stack, msg, stream, headers):
    if self._size == 0:
      channel = self._no_members
      self.__varz.no_members()
    else:
      with self._heap_lock:
        n = self.__Get()
        n.load += 1
        Heap.FixDown(self._heap, n.index, self._size)
        self._OnGet(n)

      put_called = [False]
      def PutWrapper():
        if not put_called[0]:
          with self._heap_lock:
            put_called[0] = True
            self.__Put(n)
      sink_stack.Push(self, PutWrapper)

      msg.properties[MessageProperties.Endpoint] = n.endpoint
      channel = n.channel

    channel.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    context()
    sink_stack.AsyncProcessResponse(stream, msg)

  # Events override by subclasses
  def _OnNodeDown(self, node):
    return AsyncResult.Complete()

  def _OnNodeUp(self, node):
    pass

  def _OnGet(self, node):
    pass

  def _OnPut(self, node):
    pass

  def __Get(self):
    """Get the least-loaded node from the heap.

    Returns:
      A node.
    """
    while True:
      n = self._downq
      m = None
      while n is not None:
        if n.index < 0:
          # The node has been discarded.
          n = n.downq
          if m is None:
            self._downq = n
          else:
            m.downq = n
        elif n.channel.state == ChannelState.Open:
          # The node was resurrected, mark it back up
          n.load -= self.Penalty
          ep = n.endpoint
          Heap.FixUp(self._heap, n.index)
          o = n.downq
          n.downq = None
          n = o
          if m is None:
            self._downq = n
          else:
            m.downq = n
          self._log.info('Marking node %s up' % str(ep))
        else:
          # No change, move to the next node in the linked list
          m, n = n, n.downq

      n = self._heap[1]
      if n.channel.state == ChannelState.Open or n.load >= 0:
        return n
      else:
        # Node is now down
        n.downq = self._downq
        self._downq = n
        n.load += self.Penalty
        Heap.FixDown(self._heap, 1, self._size)
        self._OnNodeDown(n)
        self._log.warning('Marking node %s down' % str(n.endpoint))
        # Loop

  def __Put(self, n):
    """'Return' a member to the heap.  Load on the node is decremented and its
    position in the heap is adjusted.

    Args:
      n - The node to return
    """
    n.load -= 1
    if n.load < self.Idle:
      self._log.warning('Decrementing load below Zero')
      n.load = self.Idle
    if n.index < 0 and n.load > self.Idle:
      pass
    elif n.index < 0 and n.load == self.Idle:
      n.channel.Close()
    elif n.load == self.Idle and self._size > 1:
      i = n.index
      Heap.Swap(self._heap, i, self._size)
      Heap.FixDown(self._heap, i, self._size - 1)

      j = random.randint(1, self._size)
      Heap.Swap(self._heap, j, self._size)
      Heap.FixUp(self._heap, j)
      Heap.FixUp(self._heap, self._size)
    else:
      Heap.FixUp(self._heap, n.index)
    self._OnPut(n)

  @synchronized
  def _AddSink(self, endpoint, sink_factory):
    """Add a sink to the heap.
    The sink is immediately opened and initialized to Zero load.

    Args:
      sink - The sink that was just added.
    """
    self._size += 1
    self.__varz.size(self._size)
    new_node = self.Node(sink_factory(), self.Idle, self._size, endpoint)
    self._heap.append(new_node)
    Heap.FixUp(self._heap, self._size)
    # Adding an Open() in here allows us to optimistically assume it'll be opened
    # before the next message attempts to get it.  However, the Open() will likely
    # yield, so other code paths need to be aware there is a potentially un-open
    # sink on the heap.
    if self._open:
      return self._OpenNode(new_node)
    else:
      return AsyncResult.Complete()

  def _FindNodeByEndpoint(self, endpoint):
    i = next((idx for idx, node in enumerate(self._heap)
              if node.endpoint == endpoint), 0)
    if i == 0:
      return None
    return self._heap[i]

  @synchronized
  def _RemoveSink(self, endpoint):
    """Remove a sink from the heap.
    The sink is closed immediately if it has no outstanding load, otherwise the
    close is deferred until the sink goes idle.

    Args:
      sink - The sink to be removed.
    """
    node = self._FindNodeByEndpoint(endpoint)
    if not node or node.index < 0:
      return False

    i = node.index
    Heap.Swap(self._heap, i, self._size)
    Heap.FixDown(self._heap, i, self._size - 1)
    self._heap.pop()
    self._size -= 1
    self.__varz.size(self._size)
    if self._size < 0:
      self._log.warning("Decrementing size below 0")
      self._size = 0
    node.index = -1
    if node.load == self.Idle or node.load >= 0:
      node.channel.Close()
    return True

  def _OnServersChanged(self, endpoint, channel_factory, added):
    """Invoked by the LoadBalancer when an endpoint joins or leaves the
    server set.

    Args:
      endpoint - A tuple of (endpoint, sink).
      added - True if the endpoint is being added, False if being removed.
    """
    if added:
      self._AddSink(endpoint, channel_factory)
    else:
      self._RemoveSink(endpoint)

  def _OnOpenNodeComplete(self, ar, node):
    if ar.exception:
      self._log.error('Exception caught opening channel: %s' % str(ar.exception))
      return self._OnNodeDown(node)
    else:
      return ar

  def _OpenNode(self, n):
    return n.channel \
      .Open() \
      .ContinueWith(lambda ar: self._OnOpenNodeComplete(ar, n)) \
      .Unwrap()

  def _OpenInitialChannels(self):
    """Open the sink and all underlying nodes."""
    self._open = True
    if self._size > 0:
      # Ignore the first sink, it's the FailingChannelSink.
      AsyncResult.WhenAny([self._OpenNode(n) for n in self._heap[1:]])\
        .ContinueWith(lambda _ar: self._OnOpenComplete())
    else:
      self._OnOpenComplete()

  def Close(self):
    """Close the sink and all underlying nodes immediately."""
    super(HeapBalancerSink, self).Close()
    self._open = False
    [n.channel.Close() for n in self._heap]

  @property
  def state(self):
    return max([n.channel.state for n in self._heap])

HeapBalancerSink.Builder = SinkProvider(
  HeapBalancerSink,
  SinkRole.LoadBalancer,
  server_set_provider = None)



from abc import abstractmethod

import logging
import time
from struct import unpack

import gevent
from gevent.queue import Queue

from ..asynchronous import (
  AsyncResult,
  NamedGreenlet
)
from ..compat import BytesIO
from ..constants import ChannelState
from ..message import (
  Deadline,
  ClientError,
  MethodReturnMessage
)
from ..sink import (
  ClientMessageSink,
)
from ..varz import (
  AggregateTimer,
  AverageTimer,
  Counter,
  Gauge,
  Rate,
  Source,
  VarzBase
)
from ..constants import TransportHeaders

ROOT_LOG = logging.getLogger('scales.mux')

class Tag(object):
  __slots__ = ('_tag',)

  KEY = "__Tag"

  def __init__(self, tag):
    self._tag = tag

  def Encode(self):
    return [self._tag >> 16 & 0xff,
            self._tag >>  8 & 0xff,
            self._tag       & 0xff]


class TagPool(object):
  """A class which manages a pool of tags
    """
  POOL_LOGGER = ROOT_LOG.getChild('TagPool')

  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.mux.TagPool'
    _VARZ = {
      'pool_exhausted': Counter,
      'max_tag': Gauge
    }

  def __init__(self, max_tag, service, host):
    self._set = set()
    self._next = 1
    self._max_tag = max_tag
    self._varz = self.Varz(Source(service=service, endpoint=host))
    self._log = self.POOL_LOGGER.getChild('[%s.%s]' % (service, host))

  def get(self):
    """Get a tag from the pool.

    Returns:
      A tag

    Raises:
      Exception if the next tag will be > max_tag
    """
    if not self._set:
      if self._next == self._max_tag - 1:
        self._varz.pool_exhausted()
        raise Exception("No tags left in pool.")
      self._next += 1
      ret_tag = self._next
      self._log.debug('Allocating new tag, max is now %d' % ret_tag)
      self._varz.max_tag(ret_tag)
      return ret_tag
    else:
      return self._set.pop()

  def release(self, tag):
    """Return a tag to the pool.

    Args:
      tag - The previously leased tag.
    """
    if tag in self._set:
      self._log.warning('Tag %d has been returned more than once!' % tag)
    self._set.add(tag)


class MuxSocketTransportSink(ClientMessageSink):
  """A transport sink for tmux servers.
  This sink supports concurrent requests over its transport.
  """

  SINK_LOG = ROOT_LOG.getChild('SocketTransportSink')
  _EMPTY_DCT = {}
  _CLOSE_INVOKED = "Close invoked"

  class Varz(VarzBase):
    """
    messages_sent - The number of messages sent over this sink.
    messages_recv - The number of messages received over this sink.
    active - 1 if the sink is open, else 0.
    send_queue_size - The length of the send queue.
    send_time - The aggregate amount of time spent sending data.
    recv_time - The aggregate amount of time spend receiving data.
    send_latency - The average amount of time taken to send a message.
    recv_latency - The average amount of time taken to receive a message
                   (once a response has reached the client).
    transport_latency - The average amount of time taken to perform a full
                        method call transaction (send data, wait for response,
                        read response).
    """
    _VARZ_BASE_NAME = 'scales.thriftmux.SocketTransportSink'
    _VARZ = {
      'messages_sent': Rate,
      'messages_recv': Rate,
      'active': Gauge,
      'send_queue_size': Gauge,
      'send_time': AggregateTimer,
      'recv_time': AggregateTimer,
      'send_latency': AverageTimer,
      'recv_latency': AverageTimer,
      'transport_latency': AverageTimer
    }

  def __init__(self, socket, service):
    super(MuxSocketTransportSink, self).__init__()
    self._socket = socket
    self._state = ChannelState.Idle
    self._log = self.SINK_LOG.getChild('[%s.%s:%d]' % (
      service, self._socket.host, self._socket.port))
    self._socket_source = '%s:%d' % (self._socket.host, self._socket.port)
    self._service = service
    self._open_result = None
    self._varz = self.Varz(Source(service=self._service,
      endpoint=self._socket_source))

  def _Init(self):
    self._tag_map = {}
    self._open_result = None
    self._tag_pool = TagPool((2 ** 24) - 1, self._service, self._socket_source)
    self._greenlets = []
    self._send_queue = Queue()

  @property
  def isActive(self):
    return self._state != ChannelState.Closed

  @property
  def state(self):
    return self._state

  def Open(self):
    """Initializes the dispatcher, opening a connection to the remote host.
    This method may only be called once.
    """
    if not self._open_result:
      self._Init()
      self._open_result = AsyncResult()
      self._open_result.SafeLink(self._OpenImpl)
    return self._open_result

  def _SpawnNamedGreenlet(self, name, *args, **kwargs):
    return NamedGreenlet.spawn(
      'Scales %s for %s [%s]' % (name, self._service, self._socket_source),
      *args,
      **kwargs)

  def _OpenImpl(self):
    try:
      self._log.debug('Opening transport.')
      self._socket.open()
      self._greenlets.append(self._SpawnNamedGreenlet('Recv Loop', self._RecvLoop))
      self._greenlets.append(self._SpawnNamedGreenlet('Send Loop', self._SendLoop))

      self._CheckInitialConnection()
      self._log.debug('Open successful')
      self._state = ChannelState.Open
      self._varz.active(1)
    except Exception as e:
      self._log.error('Exception opening socket')
      self._open_result.set_exception(e)
      self._Shutdown('Open failed')
      raise

  @abstractmethod
  def _CheckInitialConnection(self):
    raise NotImplementedError()

  def Close(self):
    self._Shutdown(self._CLOSE_INVOKED, False)

  def _Shutdown(self, reason, fault=True):
    if not self.isActive:
      return

    self._state = ChannelState.Closed

    if reason == self._CLOSE_INVOKED:
      log_fn = self._log.debug
    else:
      log_fn = self._log.warning
    log_fn('Shutting down transport [%s].' % str(reason))
    self._varz.active(0)
    self._socket.close()
    [g.kill(block=False) for g in self._greenlets]
    self._greenlets = []

    if not isinstance(reason, Exception):
      reason = Exception(str(reason))
    if fault:
      self.on_faulted.Set(reason)
    msg = MethodReturnMessage(error=ClientError(reason))

    for sink_stack, _, _ in self._tag_map.values():
      sink_stack.AsyncProcessResponseMessage(msg)

    if (self._open_result
        and not self._open_result.ready()
        and isinstance(reason, Exception)):
      self._open_result.set_exception(reason)

    self._tag_map = {}
    self._open_result = None
    self._send_queue = Queue()

  @abstractmethod
  def _OnTimeout(self, tag):
    raise NotImplementedError()

  def _HandleTimeout(self, msg_properties):
    """Determine if a message has timed out yet (because it waited in the queue
    for too long).  If it hasn't, initialize the timeout handler to fire if the
    message times out in transit.

    Args:
      msg_properties - The properties of the message.
    """
    timeout_event = msg_properties.get(Deadline.EVENT_KEY, None)
    if timeout_event and timeout_event.Get():
      # The message has timed out before it got out of the send queue
      # In this case, we can discard it immediately without even sending it.
      tag = msg_properties.pop(Tag.KEY, 0)
      if tag != 0:
        self._ReleaseTag(tag)
      return True
    elif timeout_event:
      # The event exists but hasn't been signaled yet, hook up a
      # callback so we can be notified on a timeout.
      def timeout_proc():
        timeout_tag = msg_properties.pop(Tag.KEY, 0)
        if timeout_tag:
          self._OnTimeout(timeout_tag)

      timeout_event.Subscribe(lambda evt: timeout_proc(), True)
      return False
    else:
      # No event was created, so this will never timeout.
      return False

  def _SendLoop(self):
    """Dispatch messages from the send queue to the remote server.

    Note: Messages in the queue have already been serialized into wire format.
    """
    while self.isActive:
      try:
        payload, dct = self._send_queue.get()
        queue_len = self._send_queue.qsize()
        self._varz.send_queue_size(queue_len)
        # HandleTimeout sets up the transport level timeout handling
        # for this message.  If the message times out in transit, this
        # transport will handle sending a Tdiscarded to the server.
        if self._HandleTimeout(dct): continue

        with self._varz.send_time.Measure():
          with self._varz.send_latency.Measure():
            self._socket.write(payload)
        self._varz.messages_sent()
      except Exception as e:
        self._Shutdown(e)
        break

  def _RecvLoop(self):
    """Dispatch messages from the remote server to their recipient.

    Note: Deserialization and dispatch occurs on a seperate greenlet, this only
    reads the message off the wire.
    """
    while self.isActive:
      try:
        sz, = unpack('!i', self._socket.readAll(4))
        with self._varz.recv_time.Measure():
          with self._varz.recv_latency.Measure():
            buf = BytesIO(self._socket.readAll(sz))
        self._varz.messages_recv()
        gevent.spawn(self._ProcessReply, buf)
      except Exception as e:
        self._Shutdown(e)
        break

  @abstractmethod
  def _ProcessReply(self, stream):
    raise NotImplementedError()

  def _ProcessTaggedReply(self, tag, stream):
    tup = self._ReleaseTag(tag)
    if tup:
      reply_stack, start_time, props = tup
      props[Tag.KEY] = None
      self._varz.transport_latency(time.time() - start_time)
      stream.seek(0)
      reply_stack.AsyncProcessResponseStream(stream)

  def _ReleaseTag(self, tag):
    """Return a tag to the tag pool.

    Note: Tags are only returned when the server has ACK'd them (or NACK'd) with
    and Rdispatch message (or similar).  Client initiated timeouts do NOT return
    tags to the pool.

    Args:
      tag - The tag to return.

    Returns:
      The ClientChannelSinkStack associated with the tag's response.
    """
    tup = self._tag_map.pop(tag, None)
    self._tag_pool.release(tag)
    return tup

  @abstractmethod
  def _BuildHeader(self, tag, msg_type, data_len):
    raise NotImplementedError()

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    if self._state == ChannelState.Idle and self._open_result:
      self._log.debug('Waiting for channel to be open')
      self._open_result.wait()

    if ((self._state == ChannelState.Idle and not self._open_result)
        or self._state == ChannelState.Closed) and sink_stack is not None:
      err_msg = MethodReturnMessage(error=Exception('Sink not open.'))
      sink_stack.AsyncProcessResponseMessage(err_msg)
      return

    if not msg.is_one_way:
      tag = self._tag_pool.get()
      msg.properties[Tag.KEY] = tag
      self._tag_map[tag] = (sink_stack, time.time(), msg.properties)
    else:
      tag = 0

    data_len = stream.tell()
    header = self._BuildHeader(tag, headers[TransportHeaders.MessageType], data_len)
    payload = header + stream.getvalue()
    self._send_queue.put((payload, msg.properties))

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    pass




from ..loadbalancer import ApertureBalancerSink
from ..pool import WatermarkPoolSink
from ..core import Scales

from .sink import RedisTransportSink

class Redis(object):
  @staticmethod
  def NewBuilder():
    import redis
    return Scales.NewBuilder(redis.StrictRedis) \
      .WithSink(ApertureBalancerSink.Builder()) \
      .WithSink(WatermarkPoolSink.Builder()) \
      .WithSink(RedisTransportSink.Builder())

  @staticmethod
  def NewClient(uri):
    return Redis.NewBuilder() \
      .SetUri(uri) \
      .Build()



from ..sink import SharedSinkProvider
from .sink import (
  KafkaRouterSink,
  KafkaSerializerSink,
  KafkaTransportSink
)
from ..core import Scales
from ..loadbalancer import HeapBalancerSink
from ..resurrector import ResurrectorSink


class _KafkaIface(object):
  def Put(self, topic, payloads=[], acks=1):
    pass


class Kafka(object):
  @staticmethod
  def _get_sink_key(properties):
    ep = properties['endpoint']
    label = properties['label']
    return ep.host, ep.port, label

  @staticmethod
  def NewBuilder():
    return Scales.NewBuilder(_KafkaIface) \
      .WithSink(KafkaRouterSink.Builder()) \
      .WithSink(HeapBalancerSink.Builder()) \
      .WithSink(KafkaSerializerSink.Builder()) \
      .WithSink(SharedSinkProvider(Kafka._get_sink_key)) \
      .WithSink(ResurrectorSink.Builder()) \
      .WithSink(KafkaTransportSink.Builder())

  @staticmethod
  def NewClient(leaders):
    return Kafka.NewBuilder() \
      .SetUri(leaders) \
      .Build()



"""Scales core module."""

import collections
import functools
import inspect

from six.moves.urllib.parse import urlparse, ParseResult

from .constants import (SinkProperties, SinkRole)
from .dispatch import MessageDispatcher
from .loadbalancer.serverset import (
  StaticServerSetProvider,
  ZooKeeperServerSetProvider,
)
from .sink import TimeoutSinkProvider

class _ProxyBase(object):
  def __init__(self, dispatcher):
    self._dispatcher = dispatcher
    super(_ProxyBase, self).__init__()

  def __del__(self):
    try:
      self.DispatcherClose()
    except:
      pass

  def DispatcherOpen(self):
    return self._dispatcher.Open()

  def DispatcherClose(self):
    self._dispatcher.Close()


class ClientProxyBuilder(object):
  """A class to generate proxy objects that can delegate calls to a
  message dispatcher.
  """
  _PROXY_TYPE_CACHE = {}

  @staticmethod
  def _method_name(m):
    if hasattr(m, 'func_name'):
      return m.func_name
    else:
      return m.__name__

  @staticmethod
  def _BuildServiceProxy(Iface):
    """Build a proxy class that intercepts all user methods on [Iface]
    and routes them to a message dispatcher.

    Args:
      Iface - An interface to proxy
    """
    def ProxyMethod(method_name, orig_method, asynchronous=False):
      @functools.wraps(orig_method)
      def _ProxyMethod(self, *args, **kwargs):
        ar = self._dispatcher.DispatchMethodCall(method_name, args, kwargs)
        return ar if asynchronous else ar.get()
      return _ProxyMethod

    def is_user_method(m):
      return ((inspect.ismethod(m) or inspect.isfunction(m))
              and not inspect.isbuiltin(m)
              and not ClientProxyBuilder._method_name(m).startswith('__')
              and not ClientProxyBuilder._method_name(m).endswith('__'))

    # Get all methods defined on the interface.
    iface_methods = { m[0]: ProxyMethod(*m)
                      for m in inspect.getmembers(Iface, is_user_method) }
    iface_methods.pop('__init__', None)
    iface_methods.update({ m[0] + "_async": ProxyMethod(*m, asynchronous=True)
                           for m in inspect.getmembers(Iface, is_user_method) })

    # Create a proxy class to intercept the interface's methods.
    proxy = type(
      '_ScalesTransparentProxy<%s>' % Iface.__module__,
      (_ProxyBase, Iface),
      iface_methods)
    return proxy

  @staticmethod
  def CreateServiceClient(Iface):
    """Creates a proxy class that takes all method on Iface and proxies them to
    a dispatcher.

    Args:
      Iface - A class object implementing one or more thrift interfaces.
    """
    proxy_cls = ClientProxyBuilder._PROXY_TYPE_CACHE.get(Iface, None)
    if not proxy_cls:
      proxy_cls = ClientProxyBuilder._BuildServiceProxy(Iface)
      ClientProxyBuilder._PROXY_TYPE_CACHE[Iface] = proxy_cls
    return proxy_cls


class ScalesUriParser(object):
  """Default Uri Parser for scales.  Handles both tcp:// and zk:// URIs"""
  from .loadbalancer.zookeeper import Endpoint
  Server = collections.namedtuple('Server', 'service_endpoint')

  def __init__(self):
    self.handlers = {
      'tcp': self._HandleTcp,
      'zk': self._HandleZooKeeper
    }

  def Parse(self, uri):
    parsed = urlparse(uri)
    # Work around python 2.7.3 not handling # in URIs correctly.
    if '#' in parsed.path:
      path, fragment = parsed.path.split('#', 1)
      parsed = ParseResult(
          scheme = parsed.scheme,
          netloc = parsed.netloc,
          path = path,
          params = parsed.params,
          query = parsed.query,
          fragment = fragment
      )

    handler = self.handlers.get(parsed.scheme.lower(), None)
    if not handler:
      raise Exception("No handler found for prefix %s" % parsed.scheme)
    return handler(parsed)

  def _HandleTcp(self, uri):
    servers = uri.netloc.split(',')
    server_objs = []
    for s in servers:
      host, port = s.split(':')
      server = self.Server(self.Endpoint(host, int(port)))
      server_objs.append(server)
    return StaticServerSetProvider(server_objs)

  def _HandleZooKeeper(self, uri):
    hosts = uri.netloc
    path = uri.path
    endpoint = uri.fragment if uri.fragment else None
    return ZooKeeperServerSetProvider(hosts, path, endpoint_name=endpoint)


class Scales(object):
  """Factory for scales clients."""

  SERVICE_REGISTRY = {}

  class ClientBuilder(object):
    """Builder for creating Scales clients."""

    def __init__(self, Iface):
      self._service = Iface
      self._name = Iface.__module__
      self._uri_parser = ScalesUriParser()
      self._uri = None
      self._timeout = 10
      self._server_set_provider = None
      self._client_provider = ClientProxyBuilder()
      self._properties = {}
      self._stack = []
      self._open_timeout = None

    @property
    def name(self):
      return self._name

    @property
    def server_set_provider(self):
      return self._server_set_provider

    def SetUri(self, uri):
      """Sets the URI for this client.

      By default, uri may be in the form of:
        tcp://<host:port>[,...]
      or
        zk://<zk_host:zk:port>[,...]/full/zk/path
      """
      self._uri = uri
      self._server_set_provider = self._uri_parser.Parse(uri)
      return self

    def SetUriParser(self, parser):
      """Sets the URI parser for this builder.

      Uri parsers build a ServerSetProvider from a uri."""
      self._uri_parser = parser
      return self

    def SetTimeout(self, timeout):
      """Sets the default call timeout.

      Args:
        timeout - A timeout in seconds.
      """
      self._timeout = timeout
      return self

    def SetOpenTimeout(self, timeout):
      """Sets the open timeout.

      Args:
        timeout - A timeout in seconds.  If 0, there is no wait, if None, the
                  wait is infinite.
      """
      self._open_timeout = timeout
      return self

    def SetClientProvider(self, client_provider):
      """Sets the client provider.  Client providers are used to create the
      client proxy class returned from build().

      Args:
        client_provider - An instance of a ClientProxyBuilder.
      """
      self._client_provider = client_provider
      return self

    def SetServerSetProvider(self, server_set_provider):
      """Sets the server set provider.  Must be called after SetUri(), or else
      SetUri() will overwrite the value set here.

      Args:
        server_set_provider - A subclass of ServerSetProvider.
      """

      self._server_set_provider = server_set_provider
      return self

    def SetName(self, name):
      """Sets the name of the service.  The name is used in logging and varz.

      Args:
        name - The friendly service name.
      """
      self._name = name
      return self

    def WithSink(self, sink_params):
      self._stack.append(sink_params)
      return self

    def InsertSink(self, idx, sink_params):
      self._stack.insert(idx, sink_params)
      return self

    def _Replace(self, sink_params, predicate):
      replace_idx = next((i for i, n in enumerate(self._stack)
                          if predicate(n)), -1)
      if replace_idx == -1:
        raise Exception("Unable to find sink to replace.")
      else:
        self._stack[replace_idx] = sink_params
      return self

    def ReplaceSink(self, cls_to_replace, sink_params):
      return self._Replace(sink_params, lambda n: isinstance(n, cls_to_replace))

    def ReplaceRole(self, role, sink_params):
      return self._Replace(sink_params, lambda n: n.Role == role)

    def Build(self):
      """Build a client given the current builder configuration.

      Returns:
        A proxy object with all methods of Iface.
      """
      for n, p in enumerate(self._stack[:-1]):
        p.next_provider = self._stack[n+1]

      # Add the server set provider to the properties of the load balancer in the
      # stack (if one exists)
      for n, p in enumerate(self._stack):
        if p.Role == SinkRole.LoadBalancer and p.sink_properties is not None:
          if hasattr(p.sink_properties, '_asdict'):
            new_params_dct = p.sink_properties._asdict().copy()
          elif hasattr(p.sink_properties, '__dict__'):
            new_params_dct = p.sink_properties.__dict__.copy()
          else:
            raise AttributeError('unable to copy sink_properties')
          new_params_dct['server_set_provider'] = self._server_set_provider
          p.sink_properties = p.PARAMS_CLASS(**new_params_dct)
          break

      timeout_sink = TimeoutSinkProvider()
      timeout_sink.next_provider = self._stack[0]

      properties = {
        SinkProperties.Label: self.name,
        SinkProperties.ServiceInterface: self._service,
      }
      properties.update(self._properties)

      Scales.SERVICE_REGISTRY[self.name] = (timeout_sink, properties)
      dispatcher = MessageDispatcher(
          self._service,
          timeout_sink,
          self._timeout,
          properties)

      proxy_cls = self._client_provider.CreateServiceClient(self._service)
      proxy_obj = proxy_cls(dispatcher)
      evt = proxy_obj.DispatcherOpen()
      if self._open_timeout != 0:
        evt.wait(self._open_timeout)
      return proxy_obj

  @staticmethod
  def NewBuilder(Iface):
    """Creates a new client builder for a given interface.
    All methods on the interface will be proxied into the Scales dispatcher.

    Args:
      Iface - The interface to proxy.
    Returns:
      A client builder for the interface.
    """
    return Scales.ClientBuilder(Iface)
